/*******************************************************************************
* File Name: CReset_PM.c
* Version 1.80
*
* Description:
*  This file contains the setup, control, and status commands to support 
*  the component operation in the low power mode. 
*
* Note:
*
********************************************************************************
* Copyright 2015, Cypress Semiconductor Corporation.  All rights reserved.
* You may use this file only in accordance with the license, terms, conditions, 
* disclaimers, and limitations in the end user license agreement accompanying 
* the software package with which this file was provided.
*******************************************************************************/

#include "CReset.h"

/* Check for removal by optimization */
#if !defined(CReset_Sync_ctrl_reg__REMOVED)

static CReset_BACKUP_STRUCT  CReset_backup = {0u};

    
/*******************************************************************************
* Function Name: CReset_SaveConfig
********************************************************************************
*
* Summary:
*  Saves the control register value.
*
* Parameters:
*  None
*
* Return:
*  None
*
*******************************************************************************/
void CReset_SaveConfig(void) 
{
    CReset_backup.controlState = CReset_Control;
}


/*******************************************************************************
* Function Name: CReset_RestoreConfig
********************************************************************************
*
* Summary:
*  Restores the control register value.
*
* Parameters:
*  None
*
* Return:
*  None
*
*
*******************************************************************************/
void CReset_RestoreConfig(void) 
{
     CReset_Control = CReset_backup.controlState;
}


/*******************************************************************************
* Function Name: CReset_Sleep
********************************************************************************
*
* Summary:
*  Prepares the component for entering the low power mode.
*
* Parameters:
*  None
*
* Return:
*  None
*
*******************************************************************************/
void CReset_Sleep(void) 
{
    CReset_SaveConfig();
}


/*******************************************************************************
* Function Name: CReset_Wakeup
********************************************************************************
*
* Summary:
*  Restores the component after waking up from the low power mode.
*
* Parameters:
*  None
*
* Return:
*  None
*
*******************************************************************************/
void CReset_Wakeup(void)  
{
    CReset_RestoreConfig();
}

#endif /* End check for removal by optimization */


/* [] END OF FILE */
